using System;
using System.CodeDom.Compiler;
using System.Collections.Specialized;
using System.IO;
using System.Reflection;
using System.Text;
using Microsoft.CSharp;

namespace CSI
{
    internal sealed class CSI
    {
        private static string version = "2010-06-07";

        private static string frameworkNotes =
#if FRAMEWORK_VERSION_1_1
        //            1         2         3         4         5         6         7         8
        //   12345678901234567890123456789012345678901234567890123456789012345678901234567890
            "Note: This version of CSI supports an older version of .NET in which certain\n" +
            "features are unavailable. Specifically, output coloring is not supported, so\n" +
            "the -c switch has no effect.\n" +
            "\n";
#else
            "";
#endif
#if FRAMEWORK_VERSION_1_1
        private static string framework = "1.1";
#endif // FRAMEWORK_VERSION_1_1
#if FRAMEWORK_VERSION_2_0
        private static string framework = "2.0";
#endif // FRAMEWORK_VERSION_2_0
#if FRAMEWORK_VERSION_3_0
        private static string framework = "3.0";
#endif // FRAMEWORK_VERSION_3_0
#if FRAMEWORK_VERSION_3_5
        private static string framework = "3.5";
#endif // FRAMEWORK_VERSION_3_5
#if FRAMEWORK_VERSION_4_0
        private static string framework = "4.0";
#endif // FRAMEWORK_VERSION_4_0

        private static string[] commonReferences =
        {
            "System.dll",
            "System.Data.dll",
            "System.Drawing.dll",
            "System.Windows.Forms.dll",
            "System.Xml.dll",
#if FRAMEWORK_VERSION_4_0
            "WPF\\PresentationCore.dll",
            "WPF\\PresentationFramework.dll",
            "WPF\\WindowsBase.dll",
            "System.Core.dll",
            "System.Xml.Linq.dll",
            "Microsoft.CSharp.dll",
            "System.Xaml.dll",
#endif // FRAMEWORK_VERSION_4_0
        };

        private static string[] commonReferencesProgramFiles =
        {
#if FRAMEWORK_VERSION_3_0 || FRAMEWORK_VERSION_3_5
            @"Reference Assemblies\Microsoft\Framework\v3.0\PresentationCore.dll",
            @"Reference Assemblies\Microsoft\Framework\v3.0\PresentationFramework.dll",
            @"Reference Assemblies\Microsoft\Framework\v3.0\WindowsBase.dll",
#endif // FRAMEWORK_VERSION_3_0 || FRAMEWORK_VERSION_3_5
#if FRAMEWORK_VERSION_3_5
            @"Reference Assemblies\Microsoft\Framework\v3.5\System.Core.dll",
            @"Reference Assemblies\Microsoft\Framework\v3.5\System.Xml.Linq.dll",
#endif // FRAMEWORK_VERSION_3_5
        };

        private static bool colorless;

        [STAThread]
        public static int Main(string[] args)
        {
            int returnValue = int.MaxValue;
            try
            {
                // Wrap the MainBody method to catch SecurityException
                MainBody(args, ref returnValue);
            }
            catch (CSISyntaxException e)
            {
                string[] commonReferencesNoPath = new string[commonReferences.Length];
                for (int i = 0; i < commonReferences.Length; i++)
                {
                    commonReferencesNoPath[i] = Path.GetFileName(commonReferences[i]);
                }
                string[] commonReferencesProgramFilesNoPath = new string[commonReferencesProgramFiles.Length + 1];
                commonReferencesProgramFilesNoPath[0] = "";
                for (int i = 0; i < commonReferencesProgramFiles.Length; i++)
                {
                    commonReferencesProgramFilesNoPath[i + 1] = Path.GetFileName(commonReferencesProgramFiles[i]);
                }
                CSIOutput(string.Format(
                //            1         2         3         4         5         6         7         8
                //   12345678901234567890123456789012345678901234567890123456789012345678901234567890
                    "Syntax error: {0}\n" +
                    "\n" +
                    "CSI: C# Interpreter\n" +
                    "     Version {1} for .NET {2}\n" +
                    "     http://blogs.msdn.com/b/delay/\n" +
                    "\n" +
                    "Enables the use of C# as a scripting language by executing source code files\n" +
                    "directly. The source code IS the executable, so it is easy to make changes and\n" +
                    "there is no need to maintain a separate EXE file.\n" +
                    "\n" +
                    "CSI (CodeFile)+ (-d DEFINE)* (-r Reference)* (-R)? (-q)? (-c)? (-a Arguments)?\n" +
                    "   (CodeFile)+      One or more C# source code files to execute (*.cs)\n" +
                    "   (-d DEFINE)*     Zero or more symbols to #define\n" +
                    "   (-r Reference)*  Zero or more assembly files to reference (*.dll)\n" +
                    "   (-R)?            Optional 'references' switch to include common references\n" +
                    "   (-q)?            Optional 'quiet' switch to suppress unnecessary output\n" +
                    "   (-c)?            Optional 'colorless' switch to suppress output coloring\n" +
                    "   (-a Arguments)?  Zero or more optional arguments for the executing program\n" +
                    "\n" +
                    "The list of common references included by the -R switch is:\n" +
                    "   {3}{4}\n" +
                    "\n" +
                    "CSI's return code is 2147483647 if it failed to execute the program or 0 (or\n" +
                    "whatever value the executed program returned) if it executed successfully.\n" +
                    "\n" +
                    "{5}" +
                    "Examples:\n" +
                    "   CSI Example.cs\n" +
                    "   CSI Example.cs -r System.Xml.dll -a ArgA ArgB -Switch\n" +
                    "   CSI ExampleA.cs ExampleB.cs -d DEBUG -d TESTING -R\n" +
                    "",
                    e.Message, version, framework, string.Join("\n   ", commonReferencesNoPath), string.Join("\n   ", commonReferencesProgramFilesNoPath), frameworkNotes).TrimEnd('\n'));
            }
            catch (CSIException e)
            {
                CSIOutput(e.Message);
            }
            catch (System.Security.SecurityException e)
            {
                CSIOutput(string.Format(
                //            1         2         3         4         5         6         7         8
                //   12345678901234567890123456789012345678901234567890123456789012345678901234567890
                    "SecurityException: {0}\n" +
                    "\n" +
                    "Note: The most common cause of this problem is attempting to run from a network\n" +
                    "share (by default .NET does not allow this). Please copy CSI to your machine\n" +
                    "and try again.",
                    e.Message));
            }
            catch (Exception e)
            {
                CSIOutput(string.Format("Internal error: {0}", e));
            }
            return returnValue;
        }

        public static void MainBody(string[] args, ref int returnValue)
        {
            CSIVerboseOutput("Initializing...");
            bool quiet = false;
            bool defaultReferences = false;
            bool addingArguments = false;

            CSIVerboseOutput("Parsing...");
            StringCollection codeFileNames = new StringCollection();
            StringCollection defines = new StringCollection();
            StringCollection references = new StringCollection();
            StringCollection arguments = new StringCollection();
            references.Add("System.dll");
            int i = 0;
            try
            {
                for (; i < args.Length; i++)
                {
                    if (addingArguments)
                    {
                        arguments.Add(args[i]);
                    }
                    else
                    {
                        if (args[i].StartsWith("-"))
                        {
                            switch (args[i])
                            {
                                case "-a":
                                    addingArguments = true;
                                    break;
                                case "-c":
                                    colorless = true;
                                    break;
                                case "-d":
                                    i++;
                                    defines.Add(args[i]);
                                    break;
                                case "-q":
                                    quiet = true;
                                    break;
                                case "-r":
                                    i++;
                                    references.Add(args[i]);
                                    break;
                                case "-R":
                                    defaultReferences = true;
                                    break;
                                default:
                                    throw new CSISyntaxException(string.Format("Unsupported argument {0}", args[i]));
                                    // break;
                            }
                        }
                        else
                        {
                            codeFileNames.Add(args[i]);
                        }
                    }
                }
            }
            catch (IndexOutOfRangeException)
            {
                throw new CSISyntaxException(string.Format("Missing parameter for {0}", args[i-1]));
            }
            if (0 == codeFileNames.Count)
            {
                throw new CSISyntaxException("No CodeFile specified");
            }
            if (defaultReferences)
            {
                foreach (string commonReference in commonReferences)
                {
                    references.Add(commonReference);
                }
                string programFiles = Environment.GetFolderPath(Environment.SpecialFolder.ProgramFiles);
                foreach (string commonReferenceProgramFiles in commonReferencesProgramFiles)
                {
                    references.Add(Path.Combine(programFiles, commonReferenceProgramFiles));
                }
            }

            CSIVerboseOutput("Compiling...");
            CSharpCodeProvider codeProvider = new CSharpCodeProvider(
#if FRAMEWORK_VERSION_3_5
                new System.Collections.Generic.Dictionary<string, string> { { "CompilerVersion", "v3.5" } }
#endif // FRAMEWORK_VERSION_3_5
            );
            CompilerParameters compilerParameters = new CompilerParameters();
            StringBuilder compilerOptions = new StringBuilder();
            foreach (string define in defines)
            {
                compilerOptions.AppendFormat("/define:{0} ", define);
            }
            compilerParameters.CompilerOptions = compilerOptions.ToString();
            compilerParameters.GenerateExecutable = true;
            compilerParameters.GenerateInMemory = true;
            // compilerParameters.IncludeDebugInformation = true;
            compilerParameters.WarningLevel = 4;
            foreach (string reference in references)
            {
                compilerParameters.ReferencedAssemblies.Add(reference);
            }
            string[] codeFileNamesArray = new string[codeFileNames.Count];
            codeFileNames.CopyTo(codeFileNamesArray, 0);
            CompilerResults compilerResults = 
#if FRAMEWORK_VERSION_1_1
                codeProvider.CreateCompiler().CompileAssemblyFromFileBatch
#else // FRAMEWORK_VERSION_1_1
                codeProvider.CompileAssemblyFromFile
#endif // FRAMEWORK_VERSION_1_1
                (compilerParameters, codeFileNamesArray);

            CSIVerboseOutput("Reporting...");
            bool abort = false;
            foreach (CompilerError compilerError in compilerResults.Errors)
            {
                abort |= !compilerError.IsWarning;
                if (!quiet || !compilerError.IsWarning)
                {
                    using (CreateConsoleOutputRegion((compilerError.IsWarning) ? ConsoleColor.Yellow : ConsoleColor.Red))
                    {
                        Console.WriteLine(compilerError);
                    }
                }
            }
            if (abort)
            {
                throw new CSIException("Compile error(s); aborting.");
            }

            CSIVerboseOutput("Executing...");
            string[] argumentsArray = new string[arguments.Count];
            arguments.CopyTo(argumentsArray, 0);
            try
            {
                object invokeResult;
                try
                {
                    MethodInfo entryPoint = compilerResults.CompiledAssembly.EntryPoint;
                    invokeResult = entryPoint.Invoke(null, ((0 == entryPoint.GetParameters().Length) ? null : new object[] { argumentsArray }));
                }
                finally
                {
                    returnValue = 0;
                }
                try
                {
                    returnValue = (int)invokeResult;
                }
                catch (NullReferenceException)
                {
                    // Nothing to do
                }
            }
            catch (TargetInvocationException e)
            {
                if (!quiet)
                {
                    CSIOutput(string.Format("Executing code threw an unhandled exception:\n{0}", e.InnerException.ToString()));
                }
            }

            CSIVerboseOutput("Complete.");
        }

        [System.Diagnostics.Conditional("VERBOSE")]
        private static void CSIVerboseOutput(string s)
        {
            CSIOutput(s);
        }

        private static void CSIOutput(string s)
        {
            using (CreateConsoleOutputRegion(ConsoleColor.Green))
            {
                Console.WriteLine("CSI: {0}", s);
            }
        }

        private class CSIException : Exception
        {
            public CSIException(string message)
                : base(message)
            {
            }
        }

        private class CSISyntaxException : CSIException
        {
            public CSISyntaxException(string message)
                : base(message)
            {
            }
        }

        private class DefaultConsoleOutputRegion : IDisposable
        {
            public DefaultConsoleOutputRegion()
            {
            }
            public void Dispose()
            {
            }
        }

#if !FRAMEWORK_VERSION_1_1
        private class ColoredConsoleOutputRegion : IDisposable
        {
            private readonly ConsoleColor m_previousForegroundColor;
            public ColoredConsoleOutputRegion(ConsoleColor foregroundColor)
            {
                m_previousForegroundColor = Console.ForegroundColor;
                Console.ForegroundColor = foregroundColor;
            }
            public void Dispose()
            {
                Console.ForegroundColor = m_previousForegroundColor;
            }
        }
#endif // !FRAMEWORK_VERSION_1_1

        private static IDisposable CreateConsoleOutputRegion(ConsoleColor foregroundColor)
        {
#if FRAMEWORK_VERSION_1_1
            colorless = colorless;  // Suppress warning CS0169: The private field 'CSI.CSI.colorless' is never used
            return new DefaultConsoleOutputRegion();
#else // FRAMEWORK_VERSION_1_1
            return (colorless ? ((IDisposable)(new DefaultConsoleOutputRegion())) : ((IDisposable)(new ColoredConsoleOutputRegion(foregroundColor))));
#endif // FRAMEWORK_VERSION_1_1
        }

#if FRAMEWORK_VERSION_1_1
        internal sealed class ConsoleColor
        {
            public static ConsoleColor Green = null;
            public static ConsoleColor Red = null;
            public static ConsoleColor Yellow = null;
        }
#endif // FRAMEWORK_VERSION_1_1
    }
}
